@@ -60,17 +60,26 @@ char *EventDesc[] = {
   [EVENT_STOP_THREAD]         = "Stop thread",
   [EVENT_THREAD_RUNNABLE]     = "Thread runnable",
   [EVENT_MIGRATE_THREAD]      = "Migrate thread",
-  [EVENT_SHUTDOWN]            = "Shutdown",
   [EVENT_THREAD_WAKEUP]       = "Wakeup thread",
   [EVENT_THREAD_LABEL]        = "Thread label",
+  [EVENT_STARTUP]             = "Create capabilities",
+  [EVENT_CAP_CREATE]          = "Create capability",
+  [EVENT_CAP_DELETE]          = "Delete capability",
+  [EVENT_CAP_DISABLE]         = "Disable capability",
+  [EVENT_CAP_ENABLE]          = "Enable capability",
   [EVENT_GC_START]            = "Starting GC",
   [EVENT_GC_END]              = "Finished GC",
   [EVENT_REQUEST_SEQ_GC]      = "Request sequential GC",
   [EVENT_REQUEST_PAR_GC]      = "Request parallel GC",
+  [EVENT_GC_GLOBAL_SYNC]      = "Synchronise stop-the-world GC",
+  [EVENT_GC_STATS_GHC]        = "GC statistics",
+  [EVENT_HEAP_INFO_GHC]       = "Heap static parameters",
+  [EVENT_HEAP_ALLOCATED]      = "Total heap mem ever allocated",
+  [EVENT_HEAP_SIZE]           = "Current heap size",
+  [EVENT_HEAP_LIVE]           = "Current heap live data",
   [EVENT_CREATE_SPARK_THREAD] = "Create spark thread",
   [EVENT_LOG_MSG]             = "Log message",
   [EVENT_USER_MSG]            = "User message",
-  [EVENT_STARTUP]             = "Startup",
   [EVENT_GC_IDLE]             = "GC idle",
   [EVENT_GC_WORK]             = "GC working",
   [EVENT_GC_DONE]             = "GC done",
@@ -228,7 +237,10 @@ initEventLogging(void)
         // Forked process, eventlog already started by the parent
         // before fork
         event_log_pid = getpid();
-        sprintf(event_log_filename, "%s.%d.eventlog", prog, event_log_pid);
+        // We don't have a FMT* symbol for pid_t, so we go via Word64
+        // to be sure of not losing range. It would be nicer to have a
+        // FMT* symbol or similar, though.
+        sprintf(event_log_filename, "%s.%" FMT_Word64 ".eventlog", prog, (StgWord64)event_log_pid);
     }
     stgFree(prog);
 
@@ -287,7 +299,11 @@ initEventLogging(void)
                 sizeof(EventThreadID) + sizeof(StgWord16) + sizeof(EventThreadID);
             break;
 
-        case EVENT_STARTUP:         // (cap count)
+        case EVENT_STARTUP:         // (cap_count)
+        case EVENT_CAP_CREATE:      // (cap)
+        case EVENT_CAP_DELETE:      // (cap)
+        case EVENT_CAP_ENABLE:      // (cap)
+        case EVENT_CAP_DISABLE:     // (cap)
             eventTypes[t].size = sizeof(EventCapNo);
             break;
 
@@ -322,7 +338,6 @@ initEventLogging(void)
                 sizeof(EventCapNo);
             break;
 
-        case EVENT_SHUTDOWN:        // (cap)
         case EVENT_REQUEST_SEQ_GC:  // (cap)
         case EVENT_REQUEST_PAR_GC:  // (cap)
         case EVENT_GC_START:        // (cap)
@@ -330,6 +345,7 @@ initEventLogging(void)
         case EVENT_GC_IDLE:
         case EVENT_GC_WORK:
         case EVENT_GC_DONE:
+        case EVENT_GC_GLOBAL_SYNC:  // (cap)
         case EVENT_SPARK_CREATE:    // (cap)
         case EVENT_SPARK_DUD:       // (cap)
         case EVENT_SPARK_OVERFLOW:  // (cap)
@@ -352,6 +368,31 @@ initEventLogging(void)
             eventTypes[t].size = 7 * sizeof(StgWord64);
             break;
 
+        case EVENT_HEAP_ALLOCATED:    // (heap_capset, alloc_bytes)
+        case EVENT_HEAP_SIZE:         // (heap_capset, size_bytes)
+        case EVENT_HEAP_LIVE:         // (heap_capset, live_bytes)
+            eventTypes[t].size = sizeof(EventCapsetID) + sizeof(StgWord64);
+            break;
+
+        case EVENT_HEAP_INFO_GHC:     // (heap_capset, n_generations,
+                                      //  max_heap_size, alloc_area_size,
+                                      //  mblock_size, block_size)
+            eventTypes[t].size = sizeof(EventCapsetID)
+                               + sizeof(StgWord16)
+                               + sizeof(StgWord64) * 4;
+            break;
+
+        case EVENT_GC_STATS_GHC:      // (heap_capset, generation,
+                                      //  copied_bytes, slop_bytes, frag_bytes,
+                                      //  par_n_threads,
+                                      //  par_max_copied, par_tot_copied)
+            eventTypes[t].size = sizeof(EventCapsetID)
+                               + sizeof(StgWord16)
+                               + sizeof(StgWord64) * 3
+                               + sizeof(StgWord32)
+                               + sizeof(StgWord64) * 2;
+            break;
+
         case EVENT_BLOCK_MARKER:
             eventTypes[t].size = sizeof(StgWord32) + sizeof(EventTimestamp) + 
                 sizeof(EventCapNo);
@@ -519,11 +560,6 @@ postSchedEvent (Capability *cap,
         break;
     }
 
-    case EVENT_SHUTDOWN:        // (cap)
-    {
-        break;
-    }
-
     default:
         barf("postSchedEvent: unknown event tag %d", tag);
     }
@@ -588,6 +624,7 @@ postSparkCountersEvent (Capability *cap,
     }
     
     postEventHeader(eb, EVENT_SPARK_COUNTERS);
+    /* EVENT_SPARK_COUNTERS (crt,dud,ovf,cnv,gcd,fiz,rem) */
     postWord64(eb,counters.created);
     postWord64(eb,counters.dud);
     postWord64(eb,counters.overflowed);
@@ -597,6 +634,36 @@ postSparkCountersEvent (Capability *cap,
     postWord64(eb,remaining);
 }
 
+void
+postCapEvent (EventTypeNum  tag,
+              EventCapNo    capno)
+{
+    ACQUIRE_LOCK(&eventBufMutex);
+
+    if (!hasRoomForEvent(&eventBuf, tag)) {
+        // Flush event buffer to make room for new event.
+        printAndClearEventBuf(&eventBuf);
+    }
+    
+    postEventHeader(&eventBuf, tag);
+
+    switch (tag) {
+    case EVENT_CAP_CREATE:   // (cap)
+    case EVENT_CAP_DELETE:   // (cap)
+    case EVENT_CAP_ENABLE:   // (cap)
+    case EVENT_CAP_DISABLE:  // (cap)
+    {
+        postCapNo(&eventBuf,capno);
+        break;
+    }
+
+    default:
+        barf("postCapEvent: unknown event tag %d", tag);
+    }
+
+    RELEASE_LOCK(&eventBufMutex);
+}
+
 void postCapsetEvent (EventTypeNum tag,
                       EventCapsetID capset,
                       StgWord info)
@@ -752,6 +819,101 @@ void postWallClockTime (EventCapsetID capset)
     RELEASE_LOCK(&eventBufMutex);
 }
 
+/*
+ * Various GC and heap events
+ */
+void postHeapEvent (Capability    *cap,
+                    EventTypeNum   tag,
+                    EventCapsetID  heap_capset,
+                    W_           info1)
+{
+    EventsBuf *eb;
+
+    eb = &capEventBuf[cap->no];
+
+    if (!hasRoomForEvent(eb, tag)) {
+        // Flush event buffer to make room for new event.
+        printAndClearEventBuf(eb);
+    }
+    
+    postEventHeader(eb, tag);
+
+    switch (tag) {
+    case EVENT_HEAP_ALLOCATED:     // (heap_capset, alloc_bytes)
+    case EVENT_HEAP_SIZE:          // (heap_capset, size_bytes)
+    case EVENT_HEAP_LIVE:          // (heap_capset, live_bytes)
+    {
+        postCapsetID(eb, heap_capset);
+        postWord64(eb, info1 /* alloc/size/live_bytes */);
+        break;
+    }
+
+    default:
+        barf("postHeapEvent: unknown event tag %d", tag);
+    }
+}
+
+void postEventHeapInfo (EventCapsetID heap_capset,
+                        nat           gens,
+                        W_          maxHeapSize,
+                        W_          allocAreaSize,
+                        W_          mblockSize,
+                        W_          blockSize)
+{
+    ACQUIRE_LOCK(&eventBufMutex);
+
+    if (!hasRoomForEvent(&eventBuf, EVENT_HEAP_INFO_GHC)) {
+        // Flush event buffer to make room for new event.
+        printAndClearEventBuf(&eventBuf);
+    }
+
+    postEventHeader(&eventBuf, EVENT_HEAP_INFO_GHC);
+    /* EVENT_HEAP_INFO_GHC (heap_capset, n_generations,
+                            max_heap_size, alloc_area_size,
+                            mblock_size, block_size) */
+    postCapsetID(&eventBuf, heap_capset);
+    postWord16(&eventBuf, gens);
+    postWord64(&eventBuf, maxHeapSize);
+    postWord64(&eventBuf, allocAreaSize);
+    postWord64(&eventBuf, mblockSize);
+    postWord64(&eventBuf, blockSize);
+
+    RELEASE_LOCK(&eventBufMutex);
+}
+
+void postEventGcStats  (Capability    *cap,
+                        EventCapsetID  heap_capset,
+                        nat            gen,
+                        W_           copied,
+                        W_           slop,
+                        W_           fragmentation,
+                        nat            par_n_threads,
+                        W_           par_max_copied,
+                        W_           par_tot_copied)
+{
+    EventsBuf *eb;
+
+    eb = &capEventBuf[cap->no];
+
+    if (!hasRoomForEvent(eb, EVENT_GC_STATS_GHC)) {
+        // Flush event buffer to make room for new event.
+        printAndClearEventBuf(eb);
+    }
+    
+    postEventHeader(eb, EVENT_GC_STATS_GHC);
+    /* EVENT_GC_STATS_GHC (heap_capset, generation,
+                           copied_bytes, slop_bytes, frag_bytes,
+                           par_n_threads, par_max_copied, par_tot_copied) */
+    postCapsetID(eb, heap_capset);
+    postWord16(eb, gen);
+    postWord64(eb, copied);
+    postWord64(eb, slop);
+    postWord64(eb, fragmentation);
+    postWord32(eb, par_n_threads);
+    postWord64(eb, par_max_copied);
+    postWord64(eb, par_tot_copied);
+}
+
 void
 postEvent (Capability *cap, EventTypeNum tag)
 {
@@ -767,6 +929,25 @@ postEvent (Capability *cap, EventTypeNum tag)
     postEventHeader(eb, tag);
 }
 
+void
+postEventAtTimestamp (Capability *cap, EventTimestamp ts, EventTypeNum tag)
+{
+    EventsBuf *eb;
+
+    eb = &capEventBuf[cap->no];
+
+    if (!hasRoomForEvent(eb, tag)) {
+        // Flush event buffer to make room for new event.
+        printAndClearEventBuf(eb);
+    }
+
+    /* Normally we'd call postEventHeader(), but that generates its own
+       timestamp, so we go one level lower so we can write out
+       the timestamp we received as an argument. */
+    postEventTypeNum(eb, tag);
+    postWord64(eb, ts);
+}
+
 #define BUF 512
 
 void postLogMsg(EventsBuf *eb, EventTypeNum type, char *msg, va_list ap)
